/*
 * Copyright © 2021 Fachhochschule für die Wirtschaft (FHDW) Hannover
 *
 * This file is part of ipspiel23-tictactoe-core.
 *
 * ipspiel23-tictactoe-core is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * ipspiel23-tictactoe-core is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * ipspiel23-tictactoe-core. If not, see <http://www.gnu.org/licenses/>.
 */
package de.fhdw.gaming.ipspiel23.tictactoe.core.domain.impl;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Random;

import de.fhdw.gaming.core.domain.DefaultGame;
import de.fhdw.gaming.core.domain.ObserverFactoryProvider;
import de.fhdw.gaming.ipspiel23.tictactoe.core.domain.TicTacToeField;
import de.fhdw.gaming.ipspiel23.tictactoe.core.domain.TicTacToeFieldState;
import de.fhdw.gaming.ipspiel23.tictactoe.core.domain.TicTacToeGame;
import de.fhdw.gaming.ipspiel23.tictactoe.core.domain.TicTacToeMoveChecker;
import de.fhdw.gaming.ipspiel23.tictactoe.core.domain.TicTacToePlayer;
import de.fhdw.gaming.ipspiel23.tictactoe.core.domain.TicTacToeState;
import de.fhdw.gaming.ipspiel23.tictactoe.core.domain.TicTacToeStrategy;
import de.fhdw.gaming.ipspiel23.tictactoe.core.moves.TicTacToeMove;
import de.fhdw.gaming.ipspiel23.tictactoe.core.moves.factory.TicTacToeMoveFactory;
import de.fhdw.gaming.ipspiel23.tictactoe.core.moves.impl.TicTacToeDefaultMoveFactory;

/**
 * Implements the Tic Tac Toe game.
 */
final class TicTacToeGameImpl extends DefaultGame<TicTacToePlayer, TicTacToeState, TicTacToeMove, TicTacToeStrategy>
        implements TicTacToeGame {

    /**
     * The random number generator.
     */
    private static final Random RANDOM = new Random();

    /**
     * The number of rows (and columns) of the board.
     */
    private final int boardSize;
    
    /**
     * The move factory.
     */
    private final TicTacToeMoveFactory moveFactory;

    /**
     * Creates a TicTacToe game.
     *
     * @param id                        The ID of this game.
     * @param initialState              The initial state of the game.
     * @param strategies                The players' strategies.
     * @param maxComputationTimePerMove The maximum computation time per move in seconds.
     * @param moveChecker               The move checker.
     * @param observerFactoryProvider   The {@link ObserverFactoryProvider}.
     * @throws IllegalArgumentException if the player sets do not match.
     * @throws InterruptedException     if creating the game has been interrupted.
     */
    TicTacToeGameImpl(final int id, final TicTacToeState initialState, final Map<String, TicTacToeStrategy> strategies,
            final long maxComputationTimePerMove, final TicTacToeMoveChecker moveChecker,
            final ObserverFactoryProvider observerFactoryProvider)
            throws IllegalArgumentException, InterruptedException {

        super(id, initialState, strategies, maxComputationTimePerMove, moveChecker, observerFactoryProvider);
        this.boardSize = initialState.getBoard().getSize();
        this.moveFactory = new TicTacToeDefaultMoveFactory();
    }

    @Override
    public Optional<TicTacToeMove> chooseRandomMove(final TicTacToePlayer player, final TicTacToeState state) {
        final List<TicTacToeField> fields = new ArrayList<>(
                state.getBoard().getFieldsBeing(TicTacToeFieldState.EMPTY).values());

        if (fields.isEmpty()) {
            return Optional.empty();
        }
        final int index = RANDOM.nextInt(fields.size());
        final TicTacToeField field = fields.get(index);
        return Optional.of(this.moveFactory.createPlaceMarkMove(player.isUsingCrosses(), field.getPosition()));
    }

    @Override
    public String toString() {
        return String
                .format("TicTacToeGame[id=%d, boardSize=%d, %s]", this.getId(), this.boardSize, this.gameToString());
    }

    @Override
    public int getBoardSize() {
        return this.boardSize;
    }
}
